from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.ACLineSegmentPhase].keys())
    asset_list = list(typed_catalog[cim.ACLineSegment].keys())
    
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?Location ?phase ?sequenceNumber ?ACLineSegment ?WireInfo
        (group_concat(distinct ?Measurement; separator=";") as ?Measurements) 
        WHERE {          
          ?eq r:type cim:ACLineSegmentPhase.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
        
    # add all assets
    query_message += """               }
        VALUES ?ACLineSegment {"""
    for asset_mrid in asset_list:
        query_message += ' "%s" \n' % asset_mrid
        
    # add all attributes
    query_message += """               } 
        
        #trace back up to EquipmentContainer via ACLineSegment
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        ?line cim:Equipment.EquipmentContainer ?fdr.
        ?eq cim:ACLineSegmentPhase.ACLineSegment ?line.
        ?line cim:IdentifiedObject.mRID ?ACLineSegment.

        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.
        
        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc.
                  ?loc cim:IdentifiedObject.mRID ?Location.}

        OPTIONAL {?meas cim:Measurement.PowerSystemResource ?line.
                  ?meas cim:Measurement.phases ?measphs.
                  ?meas cim:IdentifiedObject.mRID ?meas_id.
                  ?meas a ?meas_cls.
                  bind(concat(str(?meas_id),",",strafter(str(?meas_cls),"CIM100#")) as ?Measurement).}
        
        OPTIONAL {?eq cim:ACLineSegmentPhase.phase ?phs.
                 bind(strafter(str(?phs),"SinglePhaseKind.") as ?phase)}
        OPTIONAL {?eq cim:ACLineSegmentPhase.sequenceNumber ?sequenceNumber.}
        OPTIONAL {?eq cim:ACLineSegmentPhase.WireInfo ?wire.
                  ?wire cim:IdentifiedObject.mRID ?wire_id.
                  ?wire a ?wire_cls.
                  bind(concat(str(?wire_id),",",strafter(str(?wire_cls),"CIM100#")) as ?WireInfo).}

        #FILTER regex(STR(?measphs), ?phase)

        }
        GROUP by ?mRID ?name ?Location ?phase ?sequenceNumber ?ACLineSegment ?WireInfo ?measphs
        ORDER by  ?name

        """
    return query_message


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.EnergyConsumer].keys())
    
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?BaseVoltage ?Location ?p ?q ?customerCount ?grounded ?phaseConnection ?LoadResponse
            (group_concat(distinct ?Terminal; separator=";") as ?Terminals) 
            (group_concat(distinct ?Measurement; separator=";") as ?Measurements) 
            (group_concat(distinct ?EnergyConsumerPhase; separator=";") as ?EnergyConsumerPhase)
        WHERE {          
          ?eq r:type cim:EnergyConsumer.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.

        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.

        OPTIONAL {?eq cim:ConductingEquipment.BaseVoltage ?bv.
        ?bv cim:IdentifiedObject.mRID ?BaseVoltage.}

        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc.
        ?loc cim:IdentifiedObject.mRID ?Location.}

        ?t cim:Terminal.ConductingEquipment ?eq.
        ?t cim:IdentifiedObject.mRID ?Terminal

        OPTIONAL {?meas cim:Measurement.Terminal ?t.
        ?meas cim:IdentifiedObject.mRID ?meas_id.
        ?meas a ?meas_cls.
        bind(concat(str(?meas_id),",",strafter(str(?meas_cls),"CIM100#")) as ?Measurement).}

        OPTIONAL {?eq cim:EnergyConsumer.p ?p.}
        OPTIONAL {?eq cim:EnergyConsumer.q ?q.}
        OPTIONAL {?eq cim:EnergyConsumer.customerCount ?customerCount.}
        OPTIONAL {?eq cim:EnergyConsumer.grounded ?grounded.}
        
        OPTIONAL {?eq cim:EnergyConsumer.phaseConnection ?phs.
                  bind(strafter(str(?phs),"PhaseShuntConnectionKind.") as ?phaseConnection) }
        OPTIONAL {?eq cim:EnergyComsumer.LoadResponse ?lr.
                  ?lr cim:IdentifiedObject.mRID ?LoadResponse}
        OPTIONAL {?ecp cim:EnergyConsumerPhase.EnergyConsumer ?eq.
                  ?ecp cim:IdentifiedObject.mRID ?EnergyConsumerPhase.}

        }
        GROUP by ?mRID ?name ?BaseVoltage ?Location ?p ?q ?customerCount ?grounded ?phaseConnection ?LoadResponse
        ORDER by  ?name
        """
    return query_message


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List



@dataclass
class Parameter:
    key: Any
    value: Any


@dataclass
class ConnectionParameters:
    parameters: List[Parameter] = field(default_factory=list)


@dataclass
class QueryResponse:
    response: Any


@dataclass
class ConnectionInterface:
    connection_params: ConnectionParameters

    def connect(self):
        raise RuntimeError("Must have implemented connect in inherited class")

    def disconnect(self):
        raise RuntimeError("Must have implemented disconnect in inherited class")

    def load_attributes(self, obj: object):
        raise RuntimeError("Must have implemented load_attributes in inherited class")

    def create_default_instances(self, feeder_mrid: str | Feeder, mrid_list: List[str]):
        raise RuntimeError("Must have implemented retrieve_instance from inherited class")

    def execute(self, query: str) -> QueryResponse:
        raise RuntimeError("Must have implemented query in the inherited class")



from typing import Dict, List

from cimgraph.models.distributed_model import DistributedModel
from cimgraph.models.model_parsers import (add_to_catalog,
                                           add_to_typed_catalog,
                                           get_all_by_type)
from cimgraph.models.secondary_area import SecondaryArea
from cimgraph.models.switch_area import SwitchArea



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ 
    Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.RatioTapChanger].keys())
    asset_list = list(typed_catalog[cim.TransformerTank].keys())


    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?BaseVoltage ?Location ?TransformerEnd ?highStep ?lowStep ?neutralStep 
        ?normalStep ?neutralU ?initialDelay ?subsequentDelay ?ltcFlag ?controlEnabled 
        ?tculControlMode ?tapChangerControl
        (group_concat(distinct ?Terminal; separator=";") as ?Terminals) 
        (group_concat(distinct ?Measurement; separator=";") as ?Measurements) 
        
        WHERE {          
          ?eq r:type cim:RatioTapChanger.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    
    # add all assets
    query_message += """               }
        VALUES ?TransformerTank {"""
    for asset_mrid in asset_list:
        query_message += ' "%s" \n' % asset_mrid
        
    # add all attributes
    query_message += """               } 
        #get feeder id from TransformerTank
        ?eq cim:RatioTapChanger.TransformerEnd ?end.
        ?end cim:TransformerTankEnd.TransformerTank ?tank.
        ?end cim:IdentifiedObject.mRID ?TransformerEnd.
        ?tank cim:IdentifiedObject.mRID ?TransformerTank.
        ?tank cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        
        OPTIONAL {?eq cim:ConductingEquipment.BaseVoltage ?bv.
        ?bv cim:IdentifiedObject.mRID ?BaseVoltage.}

        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc.
        ?loc cim:IdentifiedObject.mRID ?Location.}

        ?t cim:Terminal.ConductingEquipment ?eq.
        ?t cim:IdentifiedObject.mRID ?Terminal

        OPTIONAL {?meas cim:Measurement.PowerSystemResource ?eq.
            ?meas cim:IdentifiedObject.mRID ?meas_id.
            ?meas a ?meas_cls.
            bind(concat(str(?meas_id),",",strafter(str(?meas_cls),"CIM100#")) as ?Measurement).}

        OPTIONAL {?eq cim:TapChanger.TapChangerControl ?cntrl.
                  ?cntrl cim:IdentifiedObject.mRID ?TapChangerControl.}
                  
        OPTIONAL {?eq cim:RatioTapChanger.stepVoltageIncrement ?stepVoltageIncrement.}
        OPTIONAL {?eq RatioTapChanger.tculControlMode ?tcul.
               bind(strafter(str(?tcul),"#TransformerControlMode") as ?tculControlMode).}
               
        OPTIONAL {?eq cim:TapChanger.highStep ?highStep.}
        OPTIONAL {?eq cim:TapChanger.lowStep ?lowStep.}
        OPTIONAL {?eq cim:TapChanger.neutralStep ?neutralStep.}
        OPTIONAL {?eq cim:TapChanger.normalStep ?normalStep.}
        OPTIONAL {?eq cim:TapChanger.neutralU ?neutralU.}
        OPTIONAL {?eq cim:TapChanger.initialDelay ?initialDelay.}
        OPTIONAL {?eq cim:TapChanger.subsequentDelay ?subsequentDelay.}
        OPTIONAL {?eq cim:TapChanger.ltcFlag ?ltcFlag.}
        OPTIONAL {?eq cim:TapChanger.controlEnabled ?controlEnabled.}
        OPTIONAL {?eq cim:TapChanger.step ?step.}

        }
        GROUP by ?mRID ?name ?BaseVoltage ?Location ?TransformerEnd ?highStep ?lowStep ?neutralStep 
        ?normalStep ?neutralU ?initialDelay ?subsequentDelay ?ltcFlag ?controlEnabled 
        ?tculControlMode ?tapChangerControl

        ORDER by  ?name
        """
    return query_message


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.PowerElectronicsConnection].keys())

    
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?BaseVoltage ?Location ?maxIFault ?maxP ?maxQ ?p ?q ?ratedS ?ratedU 
        ?aggregate ?inService
        (group_concat(distinct ?Terminal; separator=";") as ?Terminals) 
        (group_concat(distinct ?Measurement; separator=";") as ?Measurements) 
        (group_concat(distinct ?PowerElectronicsU; separator=';') as ?PowerElectronicsUnit)  
        (group_concat(distinct ?PowerElectronicsConnectionPhase; separator=';') as ?PowerElectronicsConnectionPhases) 
        WHERE {          
          ?eq r:type cim:PowerElectronicsConnection.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        
        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.
        
        OPTIONAL {?eq cim:ConductingEquipment.BaseVoltage ?bv.
        ?bv cim:IdentifiedObject.mRID ?BaseVoltage.}

        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc.
        ?loc cim:IdentifiedObject.mRID ?Location.}

        ?t cim:Terminal.ConductingEquipment ?eq.
        ?t cim:IdentifiedObject.mRID ?Terminal

        OPTIONAL {?meas cim:Measurement.PowerSystemResource ?eq.
            ?meas cim:IdentifiedObject.mRID ?meas_id.
            ?meas a ?meas_cls.
            bind(concat(str(?meas_id),",",strafter(str(?meas_cls),"CIM100#")) as ?Measurement).}

        OPTIONAL {?eq cim:PowerElectronicsConnection.PowerElectronicsUnit ?pec.
                  ?pec cim:IdentifiedObject.mRID ?pec_id.
                  ?pec a ?pec_cls.
                  bind(concat(str(?pec_id),",",strafter(str(?pec_cls),"CIM100#")) as ?PowerElectronicsU).}
                  
        OPTIONAL {?eq cim:PowerElectronicsConnection.maxIFault ?maxIFault.}
        OPTIONAL {?eq cim:PowerElectronicsConnection.p ?p.}
        OPTIONAL {?eq cim:PowerElectronicsConnection.q ?q.}
        OPTIONAL {?eq cim:PowerElectronicsConnection.maxP ?maxP.}
        OPTIONAL {?eq cim:PowerElectronicsConnection.maxQ ?maxQ.}
        OPTIONAL {?eq cim:PowerElectronicsConnection.ratedS ?ratedS.}
        OPTIONAL {?eq cim:PowerElectronicsConnection.ratedU ?ratedU.}
        
        OPTIONAL {?eq cim:Equipment.aggregate ?aggregate.}
        OPTIONAL {?eq cim:Equipment.inService ?inService.}
        #OPTIONAL {?eq cim:Equipment.networkAnalysisEnable ?networkAnalysisEnabled.}
        #OPTIONAL {?eq cim:Equipment.normallyInService ?normallyInService.}
        
        OPTIONAL {?phs cim:PowerElectronicsConnectionPhase.PowerElectronicsConnection ?eq.
                  ?phs cim:IdentifiedObject.mRID ?PowerElectronicsConnectionPhase.}
                  
        }
        GROUP by ?mRID ?name ?BaseVoltage ?Location ?maxIFault ?maxP ?maxQ ?p ?q ?ratedS ?ratedU 
        ?aggregate ?inService
        ORDER by  ?name
        """
    return query_message


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        sparql_message: query string that can be used in blazegraph connection or STOMP client
    """
    
    mrid_list = list(typed_catalog[cim.ACDCTerminal].keys())

    
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name 
        WHERE {          
          ?eq r:type cim:ACDCTerminal.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_id
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.



        }
        ORDER by  ?name
        """
    return query_message


def get_class_type_sparql(feeder_mrid, mrid_list, add_prefix=True):
    prefix = """
            PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX c:  <http://iec.ch/TC57/CIM100#>
            """
    query_message = """
                SELECT ?mRID ?name ?class
                {
                  VALUES ?fdrid {"%s"}
                  VALUES ?mRID {""" % feeder_mrid
        
    if add_prefix:
        query_message = prefix+query_message
        
    for mrid in mrid_list:
        query_message += ' "%s" \n' % mrid

    query_message += """               } 
                  ?fdr c:IdentifiedObject.mRID ?fdrid.
                  ?eq c:IdentifiedObject.name ?name.
                  ?eq c:IdentifiedObject.mRID ?mRID.
                  ?eq a ?classraw.
                  bind(strafter(str(?classraw),"CIM100#") as ?class)
                }
                GROUP BY  ?mRID ?name ?class
                ORDER by  ?fdrid
                """
    return query_message


from typing import List
from dataclasses import dataclass, field

# import cim.data_profile as cim

def get_all_attributes(feeder_id: str, mrid_list: List[str]):
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name 
        WHERE {          
          ?eq r:type cim:AssetInfo.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_id
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.



        }
        ORDER by  ?name
        """
    return query_message


import re

from cimgraph.loaders import (ConnectionInterface, ConnectionParameters,
                              Parameter, QueryResponse)
from cimgraph.loaders.blazegraph.blazegraph import BlazegraphConnection
from SPARQLWrapper import JSON, POST, SPARQLWrapper

# from cimgraph.loaders.blazegraph.query_parsers import query_parser, query_list_parser



from __future__ import annotations

import json
import logging
import os
import sys
import time
from dataclasses import dataclass, field
from typing import List

import cimgraph.data_profile as cim
from cimgraph.loaders import ConnectionInterface
from cimgraph.models.model_parsers import (add_to_catalog,
                                           add_to_typed_catalog, cim_dump)
from cimgraph.models.secondary_area import SecondaryArea

_log = logging.getLogger(__name__)


@dataclass
class SwitchArea:
    area_id: str
    connection: ConnectionInterface
    addressable_equipment: dict[str, object] = field(default_factory=dict)
    unaddressable_equipment: dict[str, object] = field(default_factory=dict)
    boundary_switches: dict[str, object] = field(default_factory=dict)
    connectivity_nodes: dict[str, object] = field(default_factory=dict)
    secondary_areas: list[object] = field(default_factory=list)
    typed_catalog: dict[type, dict[str, object]] = field(default_factory=dict)

    # Initialize empty CIM objects for all equipment in switch area
    def initialize_switch_area(self, switch_msg: dict):
        self.feeder_mrid = self.area_id.split('.')[0]  # get feeder mRID from area id
        addr_equip = self.connection.create_default_instances(self.feeder_mrid, switch_msg['addressable_equipment'])
        for obj in addr_equip:
            add_to_catalog(obj, self.addressable_equipment)
            add_to_typed_catalog(obj, self.typed_catalog)
        unaddr_equip = self.connection.create_default_instances(self.feeder_mrid, switch_msg['unaddressable_equipment'])
        for obj in unaddr_equip:
            add_to_catalog(obj, self.unaddressable_equipment)
            add_to_typed_catalog(obj, self.typed_catalog)
        conn_nodes = self.connection.create_default_instances(self.feeder_mrid, switch_msg['connectivity_node'])
        for obj in conn_nodes:
            add_to_catalog(obj, self.connectivity_nodes)
            add_to_typed_catalog(obj, self.typed_catalog)
#        bound_sw = self.connection.create_default_instances(self.feeder_mrid, switch_msg['boundary_switches'])
#        for obj in bound_sw:
#            add_to_catalog(obj, self.boundary_switches)
#            add_to_typed_catalog(obj, self.typed_catalog)

        sa_index = 0
        for sec_area_msg in switch_msg['secondary_areas']:
            # Add switch area
            sec_area = SecondaryArea(self.area_id + '.' + str(sa_index), self.connection)
            sec_area.initialize_secondary_area(sec_area_msg)
            self.secondary_areas.append(sec_area)
            # Add switch area unaddressable equipment to feeder unaddressable equipment
            self.unaddressable_equipment.update(sec_area.unaddressable_equipment)
            sa_index = sa_index + 1

    def get_all_attributes(self, cim_class):
        if cim_class in self.typed_catalog:
            self.connection.get_all_attributes(self.feeder_mrid, self.typed_catalog, cim_class)
        else:
            _log.info('no instances of '+str(cim_class.__name__)+' found in catalog.')

    def get_attributes_query(self, cim_class):
        if cim_class in self.typed_catalog:
            sparql_message = self.connection.get_attributes_query(self.feeder_mrid, self.typed_catalog, cim_class)
        else:
            _log.info('no instances of '+str(cim_class.__name__)+' found in catalog.')
            sparql_message = ''
        return sparql_message

    def __dumps__(self, cim_class):
        if cim_class in self.typed_catalog:
            json_dump = cim_dump(self.typed_catalog, cim_class)
        else:
            json_dump = {}
            _log.info('no instances of '+str(cim_class.__name__)+' found in catalog.')

        return json_dump



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ 
    Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.TapChangerControl].keys())
    asset_list = list(typed_catalog[cim.TransformerTank].keys())


    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?TapChanger ?mode ?monitoredPhase ?enabled ?discrete ?targetValue ?targetDeadband
        ?lineDropCompensation ?lineDropR ?lineDropX ?reverseLineDropR ?reverseLineDropX ?limitVoltage 
        ?Terminal
        (group_concat(distinct ?Terminal; separator=";") as ?Terminals) 
        
        WHERE {          
          ?eq r:type cim:TapChangerControl.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    
    # add all assets
    query_message += """               }
        VALUES ?TransformerTank {"""
    for asset_mrid in asset_list:
        query_message += ' "%s" \n' % asset_mrid
        
    # add all attributes
    query_message += """               } 
        #get feeder id from TransformerTank
        ?tap cim:TapChanger.TapChangerControl ?eq.
        ?tap cim:RatioTapChanger.TransformerEnd ?end.
        ?tap cim:IdentifiedObject.mRID ?TapChanger.
        ?end cim:TransformerTankEnd.TransformerTank ?tank.
        ?end cim:IdentifiedObject.mRID ?TransformerEnd.
        ?tank cim:IdentifiedObject.mRID ?TransformerTank.
        ?tank cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.

        OPTIONAL {?eq cim:RegulatingControl.mode ?md
                 bind(strafter(str(?md),"#RegulatingControlModeKind") as ?mode).}
        OPTIONAL {?eq cim:RegulatingControl.Terminal ?term.
                  ?term cim:IdentifiedObject.mRID ?Terminal.}
        OPTIONAL {?eq cim:RegulatingControl.monitoredPhase ?phs.
                  bind(strafter(str(?phs),"#PhaseCode") as ?monitoredPhase).}
        OPTIONAL {?eq cim:RegulatingControl.enabled ?enabled.}
        OPTIONAL {?eq cim:RegulatingControl ?discrete.}
        OPTIONAL {?eq cim:RegulatingControl.targetValue ?targetValue.}
        OPTIONAL {?eq cim:RegulatingControl.targetDeadband ?targetDeadband.}
        OPTIONAL {?eq cim:TapChangerControl.lineDropCompensation ?lineDropCompensation.}
        OPTIONAL {?eq cim:TapChangerControl.lineDropR ?lineDropR.}
        OPTIONAL {?eq cim:TapChangerControl.lineDropX ?lineDropX.}
        OPTIONAL {?eq cim:TapChangerControl.reverseLineDropR ?reverseLineDropR.}
        OPTIONAL {?eq cim:TapChangerControl.reverseLineDropX ?reverseLineDropX.}
        OPTIONAL {?eq cim:TapChangerControl.limitVoltage ?limitVoltage.}

        }
        GROUP by ?mRID ?name ?TapChanger ?mode ?monitoredPhase ?enabled ?discrete ?targetValue ?targetDeadband
        ?lineDropCompensation ?lineDropR ?lineDropX ?reverseLineDropR ?reverseLineDropX ?limitVoltage 
        ?Terminal

        ORDER by  ?name
        """
    return query_message


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.Terminal].keys())

    
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?connected ?sequenceNumber ?OperationalLimitSet ?ConductingEquipment ?TransformerEnd
        ?ConnectivityNode ?TopologicalNode ?RegulatingControl
        (group_concat(distinct ?Measurement; separator=";") as ?Measurements) 
        WHERE {          
          ?eq r:type cim:Terminal.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        #get feeder id from connectivity node
        ?eq cim:Terminal.ConnectivityNode ?cn.
        ?cn cim:ConnectivityNode.ConnectivityNodeContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        
        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.
        
        OPTIONAL {?eq cim:ACDCTerminal.connected ?connected.}
        OPTIONAL {?eq cim:ACDCTerminal.sequenceNumber ?sequenceNumber.}
        OPTIONAL {?eq cim:ACDCTerminal.OperationalLimitSet ?oplim.
                  ?oplim cim:IdentifiedObject.mRID ?OperationalLimitSet.}
        
        OPTIONAL {?eq cim:Terminal.ConductingEquipment ?coneq.
                  ?coneq cim:IdentifiedObject.mRID ?ConductingEquipment.}
        OPTIONAL {?xfmr cim:TransformerEnd.Terminal ?eq.
                  ?xfmr cim:IdentifiedObject.mRID ?TransformerEnd.}
        
        ?cn cim:IdentifiedObject.mRID ?ConnectivityNode.
        OPTIONAL {?cn cim:ConnectivityNode.TopologicalNode ?topo.
                  ?topo cim:IdentifiedObject.mRID ?TopologicalNode.}
        
        OPTIONAL {?regcntrl cim:RegulatingControl.Terminal ?eq.
                  ?regcntrl cim:IdentifiedObject.mRID ?RegulatingControl.}
        
        OPTIONAL {?meas cim:Measurement.Terminal ?eq.
          ?meas cim:IdentifiedObject.mRID ?Measurement}

        }
        GROUP BY  ?mRID ?name ?connected ?sequenceNumber ?OperationalLimitSet ?ConductingEquipment 
        ?TransformerEnd ?ConnectivityNode ?TopologicalNode ?RegulatingControl
        ORDER by  ?name 
    """

    return query_message



from cimgraph.data_profile.rc4_2021.gridappsd_cim_profile_rc4_2021 import (
    AcceptanceTest, AccountNotification, AccumulationKind, Accumulator,
    AccumulatorLimit, AccumulatorLimitSet, AccumulatorReset, AccumulatorValue,
    ACDCTerminal, ACLineSegment, ACLineSegmentPhase, ActivePowerLimit,
    ActivityRecord, AggregateKind, AggregateScore, Agreement, AirCompressor,
    AlertTypeList, AmiBillingReadyKind, Analog, AnalogControl, AnalogLimit,
    AnalogLimitSet, AnalogValue, Analytic, AnalyticKind, AnalyticScore,
    ApparentPowerLimit, Appointment, Approver, Asset, AssetContainer,
    AssetDeployment, AssetFailureMode, AssetFailureTypeification,
    AssetFunction, AssetGroup, AssetGroupKind, AssetHazardKind,
    AssetHealthEvent, AssetInfo, AssetKind, AssetLocationHazard,
    AssetModelUsageKind, AssetOrganisationRole, AssetOwner, AssetTestLab,
    AssetTestSampleTaker, AssetUser, AsynchronousMachine,
    AsynchronousMachineKind, AtmosphericAnalog, AtmosphericAnalogKind,
    AtmosphericPhenomenon, Author, BaseFrequency, BasePower, BaseReading,
    BaseVoltage, BasicIntervalSchedule, BatteryState, BatteryUnit, Bay,
    BranchGroup, BranchGroupTerminal, Breaker, BreakerApplicationKind,
    BreakerConfiguration, BreakerFailureReasonKind, BusbarConfiguration,
    BusbarSection, BusbarSectionInfo, Bushing, BushingInfo,
    BushingInsulationKind, BusNameMarker, Cabinet, CableConstructionKind,
    CableInfo, CableOuterJacketKind, CableShieldMaterialKind, CAESPlant,
    CatalogAssetType, Channel, CIM_GridAPPS_D_RC4_2021, Clamp, CloudCondition,
    CloudKind, CogenerationPlant, CombinedCyclePlant, ComDirectionKind,
    ComFunction, Command, ComMedia, CommodityKind, ComModule, CompositeSwitch,
    ComTechnologyKind, ConcentricNeutralCableInfo, ConductingEquipment,
    Conductor, ConfigurationEvent, ConformLoad, ConformLoadGroup,
    ConformLoadSchedule, ConnectivityNode, ConnectivityNodeContainer,
    Connector, Control, ControlledAppliance, CoordinateSystem,
    CorporateStandardKind, CoverageCodeKind, Crew, CrewMember, CrewStatusKind,
    CrewType, Currency, CurrentLimit, Curve, CurveData, CurveStyle, Customer,
    CustomerAccount, CustomerAgreement, CustomerKind, CustomerNotification,
    Cut, Cyclone, DateInterval, DateTimeInterval, DayType, DecimalQuantity,
    DemandResponseProgram, DeploymentDate, DeploymentStateKind, DERCurveData,
    DERFunction, DERGroupDispatch, DERGroupForecast, DERMonitorableParameter,
    DERParameterKind, DERUnitSymbol, DiagnosisDataSet, Disconnector, Discrete,
    DiscreteCommand, DiscreteValue, DispatchablePowerCapability,
    DispatchSchedule, Document, DocumentPersonRole, DuctBank,
    EarthFaultCompensator, Earthquake, Editor, ElectronicAddress,
    EmissionAccount, EmissionCurve, EmissionType, EmissionValueSource,
    EndDevice, EndDeviceAction, EndDeviceCapability, EndDeviceControl,
    EndDeviceControlType, EndDeviceEvent, EndDeviceEventDetail,
    EndDeviceEventType, EndDeviceFunction, EndDeviceFunctionKind,
    EndDeviceGroup, EndDeviceInfo, EndDeviceTiming, EnergyArea,
    EnergyConnection, EnergyConsumer, EnergyConsumerPhase, EnergySource,
    EnergySourcePhase, EnvironmentalAlert, EnvironmentalAnalog,
    EnvironmentalCodedValue, EnvironmentalDataAuthority,
    EnvironmentalDataProvider, EnvironmentalEvent, EnvironmentalInformation,
    EnvironmentalLocationType, EnvironmentalPhenomenon,
    EnvironmentalStringMeasurement, Equipment, EquipmentContainer,
    EquipmentFault, Estimate, ExtensionItem, ExtensionsList,
    ExternalNetworkInjection, Facility, FacilityKind, FACTSDevice,
    FailureEvent, FailureIsolationMethodKind, Fault, FaultCauseType,
    FaultImpedance, Feeder, FieldDispatchHistory, FieldDispatchStep,
    FinancialInfo, Fire, FloatQuantity, Flood, FlowDirectionKind, Forecast,
    FossilFuel, FrequencyConverter, FScale, FuelAllocationSchedule, FuelType,
    Fuse, GeneratingUnit, GenUnitOpCostCurve, GenUnitOpSchedule,
    GeographicalRegion, GeosphericAnalog, GeosphericAnalogKind,
    GeosphericPhenomenon, GrossToNetActivePowerCurve, Ground,
    GroundDisconnector, GroundingImpedance, Hazard, HealthScore,
    HeatInputCurve, HeatRateCurve, House, HouseCooling, HouseHeating,
    Hurricane, HydroEnergyConversionKind, HydroGeneratingEfficiencyCurve,
    HydroGeneratingUnit, HydroPlantStorageKind, HydroPowerPlant, HydroPump,
    HydroPumpOpSchedule, HydrosphericAnalog, HydrosphericAnalogKind,
    HydrosphericPhenomenon, IdentifiedObject,
    IEEE1547AbnormalPerfomanceCategory, IEEE1547ControlSettings, IEEE1547Info,
    IEEE1547IslandingCategory, IEEE1547NormalPerformanceCategory,
    IEEE1547Setting, IEEE1547TripSettings, IncidentHazard,
    IncrementalHeatRateCurve, InflowForecast, InspectionDataSet,
    IntegerQuantity, IntensityCodeKind, InterrupterUnit, InterrupterUnitInfo,
    InterruptingMediumKind, IntervalBlock, IntervalReading, InUseDate, IOPoint,
    IrregularIntervalSchedule, IrregularTimePoint, Issuer, Joint, Jumper,
    Junction, LabTestDataSet, Landslide, LevelVsVolumeCurve, LifecycleDate,
    LightningStrike, Limit, LimitSet, Line, LinearShuntCompensator,
    LinearShuntCompensatorPhase, LineFault, LoadArea, LoadBreakSwitch,
    LoadGroup, LoadResponseCharacteristic, Location, LocationKind,
    MacroPeriodKind, MagneticStorm, Maintainer, MaintenanceDataSet,
    Manufacturer, Measurement, MeasurementKind, MeasurementValue,
    MeasurementValueQuality, MeasurementValueSource, MeasuringPeriodKind,
    Medium, MediumKind, Meter, MeterMultiplier, MeterMultiplierKind,
    MeterReading, MeterWorkTask, MetrologyRequirement, MonthDayInterval,
    MutualCoupling, Name, NamedPhenomenon, NameType, NameTypeAuthority,
    NoLoadTest, NonConformLoad, NonConformLoadGroup, NonConformLoadSchedule,
    NonlinearShuntCompensator, NonlinearShuntCompensatorPhase,
    NonlinearShuntCompensatorPhasePoint, NonlinearShuntCompensatorPoint,
    NotificationTriggerKind, NuclearGeneratingUnit, Observation,
    OilSampleLocation, OilSpecimen, OilTemperatureSource, OpenCircuitTest,
    OperatingMechanism, OperatingMechanismInfo, OperatingMechanismKind,
    OperatingParticipant, OperatingShare, OperationalLimit,
    OperationalLimitDirectionKind, OperationalLimitSet, OperationalLimitType,
    OperationPersonRole, Operator, Organisation, OrganisationRole,
    OverfrequencyTripCurve, OverfrequencyTripCurveData, OverheadWireInfo,
    OvervoltageTripCurve, OvervoltageTripCurveData, Ownership,
    PanDemandResponse, PanDisplay, PanPricing, PanPricingDetail,
    ParallelLineSegment, ParentOrganization, PendingCalculation,
    PenstockLossCurve, PerLengthImpedance, PerLengthLineParameter,
    PerLengthPhaseImpedance, PerLengthSequenceImpedance, Person, PersonRole,
    PetersenCoil, PetersenCoilModeKind, PhaseCode, PhaseConnectedFaultKind,
    PhaseImpedanceData, PhaseShuntConnectionKind, PhaseTapChanger,
    PhaseTapChangerAsymmetrical, PhaseTapChangerLinear,
    PhaseTapChangerNonLinear, PhaseTapChangerSymmetrical, PhaseTapChangerTable,
    PhaseTapChangerTablePoint, PhaseTapChangerTabular, PhenomenonTypeification,
    PhotovoltaicUnit, Plant, PNNLTroubleCallKind, PositionPoint, PowerCutZone,
    PowerElectronicsConnection, PowerElectronicsConnectionPhase,
    PowerElectronicsUnit, PowerElectronicsWindUnit, PowerSystemResource,
    PowerTransformer, PowerTransformerEnd, PowerTransformerInfo,
    PricingStructure, Priority, Procedure, ProcedureDataSet, ProcedureKind,
    ProductAssetModel, ProtectedSwitch, PSRType, Quality61850,
    RaiseLowerCommand, RandomisationKind, RationalNumber, RatioTapChanger,
    RatioTapChangerTable, RatioTapChangerTablePoint, ReactiveCapabilityCurve,
    Reading, ReadingInterharmonic, ReadingQuality, ReadingQualityType,
    ReadingReasonKind, ReadingType, Recloser, Register,
    RegularIntervalSchedule, RegularTimePoint, RegulatingCondEq,
    RegulatingControl, RegulatingControlModeKind, RegulationSchedule,
    RelativeDisplacement, RelativeDisplacementKind, ReportingCapability,
    ReportingGroup, ReportingMethodKind, ReportingSuperGroup, Reservoir,
    RevenueKind, RightOfWay, RiskScore, RiskScoreKind, RotatingMachine,
    SampleContainerType, ScaleKind, ScheduledEvent, ScheduledEventData, Seal,
    SealConditionKind, SealKind, Season, SeasonDayTypeSchedule, Sectionaliser,
    SeriesCompensator, ServiceCategory, ServiceKind, ServiceLocation,
    ServiceMultiplier, ServiceMultiplierKind, SetPoint, ShortCircuitTest,
    ShuntCompensator, ShuntCompensatorInfo, ShuntCompensatorPhase,
    ShutdownCurve, SimpleEndDeviceFunction, SinglePhaseKind, SmartInverterMode,
    SolarGeneratingUnit, SpaceAnalog, SpaceAnalogKind, SpacePhenomenon,
    Specimen, StartIgnFuelCurve, StartMainFuelCurve, StartRampCurve,
    StartupModel, StateVariable, StaticVarCompensator, StationSupply, Status,
    SteamSendoutSchedule, StreetAddress, StreetDetail, Streetlight,
    StringMeasurement, StringMeasurementValue, StringQuantity, Structure,
    StructureSupport, SubGeographicalRegion, SubLoadArea, Substation,
    SVCControlMode, SvEstVoltage, SvInjection, SvPowerFlow,
    SvShuntCompensatorSections, SvStatus, SvSwitch, SvTapStep, SvVoltage,
    Switch, SwitchInfo, SwitchOperationSummary, SwitchPhase, SwitchSchedule,
    SynchronousMachine, SynchronousMachineKind,
    SynchronousMachineOperatingMode, TailbayLossCurve, TapChanger,
    TapChangerControl, TapChangerInfo, TapChangerTablePoint,
    TapeShieldCableInfo, TapSchedule, TargetLevelSchedule, Tariff,
    TelephoneNumber, Terminal, TestDataSet, TestKind, TestMethod, TestReason,
    TestStandard, TestVariantKind, ThermalGeneratingUnit, ThermostatController,
    ThermostatControlMode, TimeInterval, TimeIntervalKind, TimePoint,
    TimeSchedule, TopologicalIsland, TopologicalNode, Tornado, TownDetail,
    TransformerApplicationKind, TransformerCoreAdmittance, TransformerEnd,
    TransformerEndInfo, TransformerFailureReasonKind, TransformerMeshImpedance,
    TransformerStarImpedance, TransformerTank, TransformerTankEnd,
    TransformerTankInfo, TransformerTest, TransmissionModeKind,
    TropicalCycloneAustralia, TroubleReportingKind, TroubleTicket, Tsunami,
    TypeificationCondition, UnderfrequencyTripCurve,
    UnderfrequencyTripCurveData, UndervoltageTripCurve,
    UndervoltageTripCurveData, UnitMultiplier, UnitSymbol, UsagePoint,
    UsagePointConnectedKind, UsagePointGroup, UsagePointLocation,
    UserAttribute, Validity, ValueAliasSet, ValueToAlias, Version,
    VolcanicAshCloud, VoltageControlZone, VoltageLevel, VoltageLimit,
    VoltVarCurve, VoltVarCurveData, VoltWattCurve, VoltWattCurveData,
    WattVarCurve, WattVarCurveData, WeatherCodeKind, Whirlpool,
    WindGeneratingUnit, WindGenUnitKind, WindingConnection, WireAssemblyInfo,
    WireInfo, WireInsulationKind, WireMaterialKind, WirePhaseInfo,
    WirePosition, WireSpacingInfo, WireUsageKind)

__all__ = [
    "ACDCTerminal",
    "ACLineSegment",
    "ACLineSegmentPhase",
    "AcceptanceTest",
    "AccountNotification",
    "AccumulationKind",
    "Accumulator",
    "AccumulatorLimit",
    "AccumulatorLimitSet",
    "AccumulatorReset",
    "AccumulatorValue",
    "ActivePowerLimit",
    "ActivityRecord",
    "AggregateKind",
    "AggregateScore",
    "Agreement",
    "AirCompressor",
    "AlertTypeList",
    "AmiBillingReadyKind",
    "Analog",
    "AnalogControl",
    "AnalogLimit",
    "AnalogLimitSet",
    "AnalogValue",
    "Analytic",
    "AnalyticKind",
    "AnalyticScore",
    "ApparentPowerLimit",
    "Appointment",
    "Approver",
    "Asset",
    "AssetContainer",
    "AssetDeployment",
    "AssetFailureTypeification",
    "AssetFailureMode",
    "AssetFunction",
    "AssetGroup",
    "AssetGroupKind",
    "AssetHazardKind",
    "AssetHealthEvent",
    "AssetInfo",
    "AssetKind",
    "AssetLocationHazard",
    "AssetModelUsageKind",
    "AssetOrganisationRole",
    "AssetOwner",
    "AssetTestLab",
    "AssetTestSampleTaker",
    "AssetUser",
    "AsynchronousMachine",
    "AsynchronousMachineKind",
    "AtmosphericAnalog",
    "AtmosphericAnalogKind",
    "AtmosphericPhenomenon",
    "Author",
    "BaseFrequency",
    "BasePower",
    "BaseReading",
    "BaseVoltage",
    "BasicIntervalSchedule",
    "BatteryState",
    "BatteryUnit",
    "Bay",
    "BranchGroup",
    "BranchGroupTerminal",
    "Breaker",
    "BreakerApplicationKind",
    "BreakerConfiguration",
    "BreakerFailureReasonKind",
    "BusNameMarker",
    "BusbarConfiguration",
    "BusbarSection",
    "BusbarSectionInfo",
    "Bushing",
    "BushingInfo",
    "BushingInsulationKind",
    "CAESPlant",
    "CIM_GridAPPS_D_RC4_2021",
    "Cabinet",
    "CableConstructionKind",
    "CableInfo",
    "CableOuterJacketKind",
    "CableShieldMaterialKind",
    "CatalogAssetType",
    "Channel",
    "Clamp",
    "TypeificationCondition",
    "CloudCondition",
    "CloudKind",
    "CogenerationPlant",
    "ComDirectionKind",
    "ComFunction",
    "ComMedia",
    "ComModule",
    "ComTechnologyKind",
    "CombinedCyclePlant",
    "Command",
    "CommodityKind",
    "CompositeSwitch",
    "ConcentricNeutralCableInfo",
    "ConductingEquipment",
    "Conductor",
    "ConfigurationEvent",
    "ConformLoad",
    "ConformLoadGroup",
    "ConformLoadSchedule",
    "ConnectivityNode",
    "ConnectivityNodeContainer",
    "Connector",
    "Control",
    "ControlledAppliance",
    "CoordinateSystem",
    "CorporateStandardKind",
    "CoverageCodeKind",
    "Crew",
    "CrewMember",
    "CrewStatusKind",
    "CrewType",
    "Currency",
    "CurrentLimit",
    "Curve",
    "CurveData",
    "CurveStyle",
    "Customer",
    "CustomerAccount",
    "CustomerAgreement",
    "CustomerKind",
    "CustomerNotification",
    "Cut",
    "Cyclone",
    "DERCurveData",
    "DERFunction",
    "DERGroupDispatch",
    "DERGroupForecast",
    "DERMonitorableParameter",
    "DERParameterKind",
    "DERUnitSymbol",
    "DateInterval",
    "DateTimeInterval",
    "DayType",
    "DecimalQuantity",
    "DemandResponseProgram",
    "DeploymentDate",
    "DeploymentStateKind",
    "DiagnosisDataSet",
    "Disconnector",
    "Discrete",
    "DiscreteCommand",
    "DiscreteValue",
    "DispatchSchedule",
    "DispatchablePowerCapability",
    "Document",
    "DocumentPersonRole",
    "DuctBank",
    "EarthFaultCompensator",
    "Earthquake",
    "Editor",
    "ElectronicAddress",
    "EmissionAccount",
    "EmissionCurve",
    "EmissionType",
    "EmissionValueSource",
    "EndDevice",
    "EndDeviceAction",
    "EndDeviceCapability",
    "EndDeviceControl",
    "EndDeviceControlType",
    "EndDeviceEvent",
    "EndDeviceEventDetail",
    "EndDeviceEventType",
    "EndDeviceFunction",
    "EndDeviceFunctionKind",
    "EndDeviceGroup",
    "EndDeviceInfo",
    "EndDeviceTiming",
    "EnergyArea",
    "EnergyConnection",
    "EnergyConsumer",
    "EnergyConsumerPhase",
    "EnergySource",
    "EnergySourcePhase",
    "EnvironmentalAlert",
    "EnvironmentalAnalog",
    "EnvironmentalCodedValue",
    "EnvironmentalDataAuthority",
    "EnvironmentalDataProvider",
    "EnvironmentalEvent",
    "EnvironmentalInformation",
    "EnvironmentalLocationType",
    "EnvironmentalPhenomenon",
    "EnvironmentalStringMeasurement",
    "Equipment",
    "EquipmentContainer",
    "EquipmentFault",
    "Estimate",
    "ExtensionItem",
    "ExtensionsList",
    "ExternalNetworkInjection",
    "FACTSDevice",
    "FScale",
    "Facility",
    "FacilityKind",
    "FailureEvent",
    "FailureIsolationMethodKind",
    "Fault",
    "FaultCauseType",
    "FaultImpedance",
    "Feeder",
    "FieldDispatchHistory",
    "FieldDispatchStep",
    "FinancialInfo",
    "Fire",
    "FloatQuantity",
    "Flood",
    "FlowDirectionKind",
    "Forecast",
    "FossilFuel",
    "FrequencyConverter",
    "FuelAllocationSchedule",
    "FuelType",
    "Fuse",
    "GenUnitOpCostCurve",
    "GenUnitOpSchedule",
    "GeneratingUnit",
    "GeographicalRegion",
    "GeosphericAnalog",
    "GeosphericAnalogKind",
    "GeosphericPhenomenon",
    "GrossToNetActivePowerCurve",
    "Ground",
    "GroundDisconnector",
    "GroundingImpedance",
    "Hazard",
    "HealthScore",
    "HeatInputCurve",
    "HeatRateCurve",
    "House",
    "HouseCooling",
    "HouseHeating",
    "Hurricane",
    "HydroEnergyConversionKind",
    "HydroGeneratingEfficiencyCurve",
    "HydroGeneratingUnit",
    "HydroPlantStorageKind",
    "HydroPowerPlant",
    "HydroPump",
    "HydroPumpOpSchedule",
    "HydrosphericAnalog",
    "HydrosphericAnalogKind",
    "HydrosphericPhenomenon",
    "IEEE1547AbnormalPerfomanceCategory",
    "IEEE1547ControlSettings",
    "IEEE1547Info",
    "IEEE1547IslandingCategory",
    "IEEE1547NormalPerformanceCategory",
    "IEEE1547Setting",
    "IEEE1547TripSettings",
    "IOPoint",
    "IdentifiedObject",
    "InUseDate",
    "IncidentHazard",
    "IncrementalHeatRateCurve",
    "InflowForecast",
    "InspectionDataSet",
    "IntegerQuantity",
    "IntensityCodeKind",
    "InterrupterUnit",
    "InterrupterUnitInfo",
    "InterruptingMediumKind",
    "IntervalBlock",
    "IntervalReading",
    "IrregularIntervalSchedule",
    "IrregularTimePoint",
    "Issuer",
    "Joint",
    "Jumper",
    "Junction",
    "LabTestDataSet",
    "LaborelecStandard",
    "LaborelecStandardEditionKind",
    "LaborelecStandardKind",
    "Landslide",
    "LevelVsVolumeCurve",
    "LifecycleDate",
    "LightningStrike",
    "Limit",
    "LimitSet",
    "Line",
    "LineFault",
    "LinearShuntCompensator",
    "LinearShuntCompensatorPhase",
    "LoadArea",
    "LoadBreakSwitch",
    "LoadGroup",
    "LoadResponseCharacteristic",
    "Location",
    "LocationKind",
    "MacroPeriodKind",
    "MagneticStorm",
    "Maintainer",
    "MaintenanceDataSet",
    "Manufacturer",
    "Measurement",
    "MeasurementKind",
    "MeasurementValue",
    "MeasurementValueQuality",
    "MeasurementValueSource",
    "MeasuringPeriodKind",
    "Medium",
    "MediumKind",
    "Meter",
    "MeterMultiplier",
    "MeterMultiplierKind",
    "MeterReading",
    "MeterWorkTask",
    "MetrologyRequirement",
    "MonthDayInterval",
    "MutualCoupling",
    "Name",
    "NameType",
    "NameTypeAuthority",
    "NamedPhenomenon",
    "NoLoadTest",
    "NonConformLoad",
    "NonConformLoadGroup",
    "NonConformLoadSchedule",
    "NonlinearShuntCompensator",
    "NonlinearShuntCompensatorPhase",
    "NonlinearShuntCompensatorPhasePoint",
    "NonlinearShuntCompensatorPoint",
    "NotificationTriggerKind",
    "NuclearGeneratingUnit",
    "Observation",
    "OilSampleLocation",
    "OilSpecimen",
    "OilTemperatureSource",
    "OpenCircuitTest",
    "OperatingMechanism",
    "OperatingMechanismInfo",
    "OperatingMechanismKind",
    "OperatingParticipant",
    "OperatingShare",
    "OperationPersonRole",
    "OperationalLimit",
    "OperationalLimitDirectionKind",
    "OperationalLimitSet",
    "OperationalLimitType",
    "Operator",
    "Organisation",
    "OrganisationRole",
    "OverfrequencyTripCurve",
    "OverfrequencyTripCurveData",
    "OverheadWireInfo",
    "OvervoltageTripCurve",
    "OvervoltageTripCurveData",
    "Ownership",
    "PNNLTroubleCallKind",
    "PSRType",
    "PanDemandResponse",
    "PanDisplay",
    "PanPricing",
    "PanPricingDetail",
    "ParallelLineSegment",
    "ParentOrganization",
    "PendingCalculation",
    "PenstockLossCurve",
    "PerLengthImpedance",
    "PerLengthLineParameter",
    "PerLengthPhaseImpedance",
    "PerLengthSequenceImpedance",
    "Person",
    "PersonRole",
    "PetersenCoil",
    "PetersenCoilModeKind",
    "PhaseCode",
    "PhaseConnectedFaultKind",
    "PhaseImpedanceData",
    "PhaseShuntConnectionKind",
    "PhaseTapChanger",
    "PhaseTapChangerAsymmetrical",
    "PhaseTapChangerLinear",
    "PhaseTapChangerNonLinear",
    "PhaseTapChangerSymmetrical",
    "PhaseTapChangerTable",
    "PhaseTapChangerTablePoint",
    "PhaseTapChangerTabular",
    "PhenomenonTypeification",
    "PhotovoltaicUnit",
    "Plant",
    "PositionPoint",
    "PowerCutZone",
    "PowerElectronicsConnection",
    "PowerElectronicsConnectionPhase",
    "PowerElectronicsUnit",
    "PowerElectronicsWindUnit",
    "PowerSystemResource",
    "PowerTransformer",
    "PowerTransformerEnd",
    "PowerTransformerInfo",
    "PricingStructure",
    "Priority",
    "Procedure",
    "ProcedureDataSet",
    "ProcedureKind",
    "ProductAssetModel",
    "ProtectedSwitch",
    "Quality61850",
    "RaiseLowerCommand",
    "RandomisationKind",
    "RatioTapChanger",
    "RatioTapChangerTable",
    "RatioTapChangerTablePoint",
    "RationalNumber",
    "ReactiveCapabilityCurve",
    "Reading",
    "ReadingInterharmonic",
    "ReadingQuality",
    "ReadingQualityType",
    "ReadingReasonKind",
    "ReadingType",
    "Recloser",
    "Register",
    "RegularIntervalSchedule",
    "RegularTimePoint",
    "RegulatingCondEq",
    "RegulatingControl",
    "RegulatingControlModeKind",
    "RegulationSchedule",
    "RelativeDisplacement",
    "RelativeDisplacementKind",
    "ReportingCapability",
    "ReportingGroup",
    "ReportingMethodKind",
    "ReportingSuperGroup",
    "Reservoir",
    "RevenueKind",
    "RightOfWay",
    "RiskScore",
    "RiskScoreKind",
    "RotatingMachine",
    "SVCControlMode",
    "SampleContainerType",
    "ScaleKind",
    "ScheduledEvent",
    "ScheduledEventData",
    "Seal",
    "SealConditionKind",
    "SealKind",
    "Season",
    "SeasonDayTypeSchedule",
    "Sectionaliser",
    "SeriesCompensator",
    "ServiceCategory",
    "ServiceKind",
    "ServiceLocation",
    "ServiceMultiplier",
    "ServiceMultiplierKind",
    "SetPoint",
    "ShortCircuitTest",
    "ShuntCompensator",
    "ShuntCompensatorInfo",
    "ShuntCompensatorPhase",
    "ShutdownCurve",
    "SimpleEndDeviceFunction",
    "SinglePhaseKind",
    "SmartInverterMode",
    "SolarGeneratingUnit",
    "SpaceAnalog",
    "SpaceAnalogKind",
    "SpacePhenomenon",
    "Specimen",
    "StartIgnFuelCurve",
    "StartMainFuelCurve",
    "StartRampCurve",
    "StartupModel",
    "StateVariable",
    "StaticVarCompensator",
    "StationSupply",
    "Status",
    "SteamSendoutSchedule",
    "StreetAddress",
    "StreetDetail",
    "Streetlight",
    "StringMeasurement",
    "StringMeasurementValue",
    "StringQuantity",
    "Structure",
    "StructureSupport",
    "SubGeographicalRegion",
    "SubLoadArea",
    "Substation",
    "SvEstVoltage",
    "SvInjection",
    "SvPowerFlow",
    "SvShuntCompensatorSections",
    "SvStatus",
    "SvSwitch",
    "SvTapStep",
    "SvVoltage",
    "Switch",
    "SwitchInfo",
    "SwitchOperationSummary",
    "SwitchPhase",
    "SwitchSchedule",
    "SynchronousMachine",
    "SynchronousMachineKind",
    "SynchronousMachineOperatingMode",
    "TailbayLossCurve",
    "TapChanger",
    "TapChangerControl",
    "TapChangerInfo",
    "TapChangerTablePoint",
    "TapSchedule",
    "TapeShieldCableInfo",
    "TargetLevelSchedule",
    "Tariff",
    "TelephoneNumber",
    "Terminal",
    "TestDataSet",
    "TestKind",
    "TestMethod",
    "TestReason",
    "TestStandard",
    "TestVariantKind",
    "ThermalGeneratingUnit",
    "ThermostatControlMode",
    "ThermostatController",
    "TimeInterval",
    "TimeIntervalKind",
    "TimePoint",
    "TimeSchedule",
    "TopologicalIsland",
    "TopologicalNode",
    "Tornado",
    "TownDetail",
    "TransformerApplicationKind",
    "TransformerCoreAdmittance",
    "TransformerEnd",
    "TransformerEndInfo",
    "TransformerFailureReasonKind",
    "TransformerMeshImpedance",
    "TransformerStarImpedance",
    "TransformerTank",
    "TransformerTankEnd",
    "TransformerTankInfo",
    "TransformerTest",
    "TransmissionModeKind",
    "TropicalCycloneAustralia",
    "TroubleReportingKind",
    "TroubleTicket",
    "Tsunami",
    "UnderfrequencyTripCurve",
    "UnderfrequencyTripCurveData",
    "UndervoltageTripCurve",
    "UndervoltageTripCurveData",
    "UnitMultiplier",
    "UnitSymbol",
    "UsagePoint",
    "UsagePointConnectedKind",
    "UsagePointGroup",
    "UsagePointLocation",
    "UserAttribute",
    "Validity",
    "ValueAliasSet",
    "ValueToAlias",
    "Version",
    "VolcanicAshCloud",
    "VoltVarCurve",
    "VoltVarCurveData",
    "VoltWattCurve",
    "VoltWattCurveData",
    "VoltageControlZone",
    "VoltageLevel",
    "VoltageLimit",
    "WattVarCurve",
    "WattVarCurveData",
    "WeatherCodeKind",
    "Whirlpool",
    "WindGenUnitKind",
    "WindGeneratingUnit",
    "WindingConnection",
    "WireAssemblyInfo",
    "WireInfo",
    "WireInsulationKind",
    "WireMaterialKind",
    "WirePhaseInfo",
    "WirePosition",
    "WireSpacingInfo",
    "WireUsageKind",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.TransformerEndInfo].keys())
    asset_list = list(typed_catalog[cim.TransformerTank].keys())
    
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?ratedS ?ratedU ?r ?connectionKind ?phaseAngleClock ?endNumber
        ?shortTermS ?emergencyS ?insulationU ?GroundedEndShortCircuitTests ?EnergisedEndShortCircuitTests
        ?EnergisedEndNoLoadTests ?FromMeshImpedance ?ToMeshImpedance ?CoreAdmittance ?TransformerTankInfo
        (group_concat(distinct ?Asset; separator=";") as ?Assets) 

        WHERE {          
          ?eq r:type cim:TransformerEndInfo.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
        
    # add all assets
    query_message += """               }
        VALUES ?TransformerTank {"""
    for asset_mrid in asset_list:
        query_message += ' "%s" \n' % asset_mrid
        
    # add all attributes
    query_message += """          }
        #get feeder id from TransformerTank
        ?eq cim:TransformerEndInfo.TransformerTankInfo ?tankinfo.
        ?asset cim:Asset.AssetInfo ?tankinfo.
        ?asset cim:Asset.PowerSystemResources ?tank.
        ?tank cim:IdentifiedObject.mRID ?TransformerTank.
        ?tank cim:Equipment.EquipmentContainer ?fdr.
        ?fdr cim:IdentifiedObject.mRID ?fdrid.

        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.
        
        ?tankinfo cim:IdentifiedObject.mRID ?TransformerTankInfo.
        ?asset cim:IdentifiedObject.mRID ?Asset
       
        OPTIONAL {?eq cim:TransformerEndInfo.endNumber ?endNumber.}
        OPTIONAL {?eq cim:TransformerEndInfo.connectionKind ?conn.
                  bind(strafter(str(?conn),"WindingConnection.") as ?connectionKind).}
        OPTIONAL {?eq cim:TransformerEndInfo.phaseAngleClock ?phaseAngleClock.}
        OPTIONAL {?eq cim:TransformerEndInfo.ratedU ?ratedU.}
        OPTIONAL {?eq cim:TransformerEndInfo.ratedS ?ratedS.}
        OPTIONAL {?eq cim:TransformerEndInfo.shortTermS ?shortTermS.}
        OPTIONAL {?eq cim:TransformerEndInfo.emergencyS ?emergencyS.}
        OPTIONAL {?eq cim:TransformerEndInfo.r ?r.}
        OPTIONAL {?eq cim:TransformerEndInfo.insulationU ?insulationU.}
        
        OPTIONAL {?gsct cim:ShortCircuitTest.GroundedEnds ?eq.
                  ?gsct cim:IdentifiedObject.mRID ?GroundedEndShortCircuitTests.}
                  
        OPTIONAL {?esct cim:ShortCircuitTest.EnergisedEnd ?eq.
                  ?esct cim:IdentifiedObject.mRID ?EnergisedEndShortCircuitTests.}
                  
        OPTIONAL {?enol cim:NoLoadTest.EnergisedEnd ?eq.
                  ?enol cim:IdentifiedObject.mRID ?EnergisedEndNoLoadTests.}
                  
        OPTIONAL {?tac cim:TransformerCoreAdmittance.TransformerEnd ?eq.
                  ?tac cim:IdentifiedObject.mRID ?CoreAdmittance.}
          
        OPTIONAL {?frmesh cim:TransformerMeshImpedance.FromTransformerEnd ?eq.
                  ?frmesh cim:IdentifiedObject.mRID ?FromMeshImpedance.}
          
        OPTIONAL {?frmesh cim:TransformerMeshImpedance.ToTransformerEnd ?eq.
                  ?frmesh cim:IdentifiedObject.mRID ?ToMeshImpedance.}
        }
     
        GROUP by ?mRID ?name ?ratedS ?ratedU ?r ?connectionKind ?phaseAngleClock ?endNumber
        ?shortTermS ?emergencyS ?insulationU ?GroundedEndShortCircuitTests ?EnergisedEndShortCircuitTests
        ?EnergisedEndNoLoadTests ?FromMeshImpedance ?ToMeshImpedance ?CoreAdmittance ?TransformerTankInfo
        ORDER by ?name
          """
    return query_message


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.OverheadWireInfo].keys())
    asset_list = list(typed_catalog[cim.ACLineSegment].keys())
    
    query_message = """
        PREFIX r: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim: <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?coreRadius ?coreStrandCount ?gmr ?insulated ?insulationMaterial 
         ?insulationThickness ?material ?rAC25 ?rAC50 ?rAC75 ?rDC20 ?radius ?ratedCurrent ?strandCount
         ?sizeDescription 
         (group_concat(distinct ?ACLineSegmentPhase; separator=';') as ?ACLineSegmentPhases)
        WHERE {          
          ?eq r:type cim:OverheadWireInfo.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
        
    # add all assets
    query_message += """               }
        VALUES ?ACLineSegment {"""
    for asset_mrid in asset_list:
        query_message += ' "%s" \n' % asset_mrid
        
    # add all attributes
    query_message += """               } 
        #trace back up to EquipmentContainer via ACLineSegment
        ?fdr cim:IdentifiedObject.mRID ?fdrid.
        ?line cim:Equipment.EquipmentContainer ?fdr.
        ?line cim:IdentifiedObject.mRID ?ACLineSegment.
        ?linephase cim:ACLineSegmentPhase.ACLineSegment ?line.
        ?linephase cim:ACLineSegmentPhase.WireInfo ?eq.
        ?linephase cim:IdentifiedObject.mRID ?ACLineSegmentPhase.

        ?eq cim:IdentifiedObject.mRID ?mRID.
        ?eq cim:IdentifiedObject.name ?name.
        
        #collect inherited WireInfo attributes
        OPTIONAL {?eq cim:WireInfo.coreRadius ?coreRadius.}
        OPTIONAL {?eq cim:WireInfo.coreStrandCount ?coreStrandCount.}        
        OPTIONAL {?eq cim:WireInfo.gmr ?gmr.}
        OPTIONAL {?eq cim:WireInfo.insulated ?insulated.} 
        OPTIONAL {?eq cim:WireInfo.insulationMaterial ?ins.
                  bind(strafter(str(?ins),"WireInsulationKind.") as ?insulationMaterial).}
        OPTIONAL {?eq cim:WireInfo.insulationThickness ?insulationThickness.}
        OPTIONAL {?eq cim:WireInfo.material ?material.}
        OPTIONAL {?eq cim:WireInfo.rAC25 ?rAC25.}
        OPTIONAL {?eq cim:WireInfo.rAC50 ?rAC50.}
        OPTIONAL {?eq cim:WireInfo.rAC75 ?rAC75.}
        OPTIONAL {?eq cim:WireInfo.rDC20 ?rDC20.}
        OPTIONAL {?eq cim:WireInfo.radius ?radius.}
        OPTIONAL {?eq cim:WireInfo.ratedCurrent ?ratedCurrent.}
        OPTIONAL {?eq cim:WireInfo.strandCount ?strandCount.}
        OPTIONAL {?eq cim:WireInfo.sizeDescription ?sizeDescription.}


        }
        GROUP BY ?mRID ?name ?coreRadius ?coreStrandCount ?gmr ?insulated ?insulationMaterial 
         ?insulationThickness ?material ?rAC25 ?rAC50 ?rAC75 ?rDC20 ?radius ?ratedCurrent ?strandCount
         ?sizeDescription
         
        ORDER by  ?name
        """
    return query_message


from __future__ import annotations

import traceback
from typing import Dict, List, Optional

from cimgraph.models.model_parsers import add_to_catalog, add_to_typed_catalog


def query_parser(conn, cim, feeder_mrid, typed_catalog:Dict, class_name:str, mRID:str, query:List, attribute:str, separator:str) -> object | str:
#     try:
    value = query[attribute]['value']

    #if attribute is CIM class, then build CIM objects. otherwise assign to obj_list
#         print(attribute)
    if attribute in cim.__all__:
        value = value.split(separator)
        result = build_cim_object(conn, cim, feeder_mrid, typed_catalog, value)
        if len(result) == 1:
            result = result[0]
    else:
        result = value
    setattr(typed_catalog[class_name][mRID], attribute, result)
#     except Exception:
#         traceback.print_exc()


def query_list_parser(conn, cim, feeder_mrid, typed_catalog:Dict, class_name:type, mRID:str, query:List, attribute:str, attribute_class:str, separator:str):
    value = query[attribute]['value']
    values = value.split(separator)
    #if attribute is CIM class, then build CIM objects. otherwise assign to obj_list
    if attribute_class in cim.__all__:
        obj_list = build_cim_object(conn, cim, feeder_mrid, typed_catalog, values)
    else:
        obj_list = values
    #set attribute of queried object to list parsed from query results
    setattr(typed_catalog[class_name][mRID], attribute, obj_list)
    

def build_cim_object(conn, cim, feeder_mrid, typed_catalog:Dict, mRID_list:List[str]) -> List(object):
    sparql_message = sparql.get_class_type_sparql(feeder_mrid, mRID_list)
    #execute sparql query
    query_output = conn.execute(sparql_message)
    # parse query results and add new CIM objects to list
    obj_list = [] 
#     print(query_output)
    for result in query_output['results']['bindings']:
       # print(result)
        class_name = result['class']['value']
        cls = class_name
        mRID = result['mRID']['value']
        name = result['name']['value']
        if len(mRID) > 0:
            try: 
                class_type = eval(f"cim.{cls}")
                #add class to typed_catalog if not already defined
                if class_type not in typed_catalog.keys():
                    typed_catalog[class_type] = {}

        #         for mRID in mRID_list:
                    #add object to typed_catalog if not already included
                if mRID in typed_catalog[class_type].keys():
                    obj = typed_catalog[class_type][mRID]
                else:
                        obj = class_type(mRID = mRID, name = name)
                        add_to_typed_catalog(obj, typed_catalog)
            except:
                obj = mRID
                print('warning: object class missing from data profile:', cls)

            obj_list.append(obj)
    return obj_list





from cimgraph.data_profile.cimext_2022.gridappsd_cim_profile import (
    AbnormalOPcatKind, ACDCTerminal, ACLineSegment, ACLineSegmentPhase,
    ActivityRecord, AmiBillingReadyKind, Asset, AssetContainer, AssetFunction,
    AssetInfo, BaseVoltage, BasicIntervalSchedule, Breaker, BusbarSection,
    BusbarSectionInfo, CableConstructionKind, CableInfo, CableOuterJacketKind,
    CableShieldMaterialKind, ConcentricNeutralCableInfo, ConductingEquipment,
    Conductor, ConfigurationEvent, ConnectivityNode, ConnectivityNodeContainer,
    Connector, ConstantPowerFactorSettingKind, ConstantPowerFactorSettings,
    ConstantReactivePowerSettings, Control, ConverterControlMode,
    CoordinateSystem, Curve, CurveData, CurveStyle, Customer, DateTimeInterval,
    DayType, DemandResponseProgram, DERCurveData, DERDynamics, DERFunction,
    DERGroupDispatch, DERGroupForecast, DERIEEEType1, DERMonitorableParameter,
    DERNameplateData, DERNameplateDataApplied, DERParameterKind, DERUnitSymbol,
    DispatchablePowerCapability, DispatchSchedule, DynamicsFunctionBlock,
    EndDevice, EndDeviceAction, EndDeviceCapability, EndDeviceControl,
    EndDeviceControlType, EndDeviceEvent, EndDeviceEventDetail,
    EndDeviceEventType, EndDeviceFunction, EndDeviceGroup, EndDeviceInfo,
    EndDeviceTiming, EnergyConnection, EnergyConsumer, EnergyConsumerPhase,
    EnergySource, EnergySourcePhase, Equipment, EquipmentContainer,
    EquipmentFault, Estimate, Fault, FaultCauseType, FaultImpedance, Feeder,
    FloatQuantity, FlowDirectionKind, FrequencyDroopSettings,
    FrequencyTripSettings, Fuse, GAPPS_CIM_export, GeographicalRegion, House,
    HouseCooling, HouseHeating, IdentifiedObject,
    IEEE1547AbnormalPerfomanceCategory, IEEE1547ControlSettings, IEEE1547Info,
    IEEE1547IslandingCategory, IEEE1547NormalPerformanceCategory,
    IEEE1547Setting, IEEE1547TripSettings, IOPoint, LinearShuntCompensator,
    LinearShuntCompensatorPhase, LoadBreakSwitch, LoadResponseCharacteristic,
    Location, Measurement, Meter, MeterReading, MomentaryCessationSettings,
    Name, NoLoadTest, NormalOPcatKind, OpenCircuitTest, OperationalLimit,
    OperationalLimitDirectionKind, OperationalLimitSet, OperationalLimitType,
    OrderedPhaseCodeKind, OverfrequencyTripCurve, OverheadWireInfo,
    OvervoltageTripCurve, PerLengthImpedance, PerLengthLineParameter,
    PerLengthSequenceImpedance, PhaseCode, PhaseConnectedFaultKind,
    PhaseShuntConnectionKind, PhaseTapChanger, PowerCutZone,
    PowerElectronicsConnection, PowerElectronicsConnectionPhase,
    PowerElectronicsUnit, PowerLimitSettings, PowerSystemResource,
    PowerTransformer, PowerTransformerEnd, PowerTransformerInfo, Priority,
    ProtectedSwitch, RandomisationKind, RatioTapChanger,
    ReactiveCapabilityCurve, Recloser, RegularIntervalSchedule,
    RegularTimePoint, RegulatingCondEq, RegulatingControl,
    RegulatingControlModeKind, RegulationSchedule, RemoteInputSignal,
    RemoteSignalKind, RotatingMachine, ScheduledEvent, Season,
    SeasonDayTypeSchedule, Sectionaliser, ServiceSettings, ShortCircuitTest,
    ShuntCompensator, ShuntCompensatorInfo, ShuntCompensatorPhase,
    SinglePhaseKind, StateVariable, Status, StringQuantity,
    SubGeographicalRegion, Substation, SvEstVoltage, SvInjection, SvPowerFlow,
    SvShuntCompensatorSections, SvStatus, SvTapStep, SvVoltage, Switch,
    SwitchInfo, SynchronousMachine, TapChanger, TapChangerControl,
    TapChangerInfo, TapeShieldCableInfo, TapSchedule, Terminal,
    ThermostatController, ThermostatControlMode, TimeIntervalKind,
    TopologicalIsland, TopologicalNode, TransformerCoreAdmittance,
    TransformerEnd, TransformerEndInfo, TransformerMeshImpedance,
    TransformerTank, TransformerTankEnd, TransformerTankInfo, TransformerTest,
    UnderfrequencyTripCurve, UndervoltageTripCurve, UnitMultiplier, UnitSymbol,
    UsagePoint, UsagePointConnectedKind, UsagePointGroup, Version,
    VoltageControlZone, VoltageTripSettings, VoltVarCurve, VoltVarSettings,
    VoltWattCurve, VoltWattSettings, WattVarCurve, WattVarSettings,
    WindingConnection, WireAssemblyInfo, WireInfo, WireInsulationKind,
    WireMaterialKind, WirePhaseInfo, WirePosition, WireSpacingInfo)

__all__ = [
    "ACDCTerminal",
    "ACLineSegment",
    "ACLineSegmentPhase",
    "AbnormalOPcatKind",
    "ActivityRecord",
    "AmiBillingReadyKind",
    "Asset",
    "AssetContainer",
    "AssetFunction",
    "AssetInfo",
    "BaseVoltage",
    "BasicIntervalSchedule",
    "Breaker",
    "BusbarSection",
    "BusbarSectionInfo",
    "CableConstructionKind",
    "CableInfo",
    "CableOuterJacketKind",
    "CableShieldMaterialKind",
    "ConcentricNeutralCableInfo",
    "ConductingEquipment",
    "Conductor",
    "ConfigurationEvent",
    "ConnectivityNode",
    "ConnectivityNodeContainer",
    "Connector",
    "ConstantPowerFactorSettingKind",
    "ConstantPowerFactorSettings",
    "ConstantReactivePowerSettings",
    "Control",
    "ConverterControlMode",
    "CoordinateSystem",
    "Curve",
    "CurveData",
    "CurveStyle",
    "Customer",
    "DERCurveData",
    "DERDynamics",
    "DERFunction",
    "DERGroupDispatch",
    "DERGroupForecast",
    "DERIEEEType1",
    "DERMonitorableParameter",
    "DERNameplateData",
    "DERNameplateDataApplied",
    "DERParameterKind",
    "DERUnitSymbol",
    "DateTimeInterval",
    "DayType",
    "DemandResponseProgram",
    "DispatchSchedule",
    "DispatchablePowerCapability",
    "DynamicsFunctionBlock",
    "EndDevice",
    "EndDeviceAction",
    "EndDeviceCapability",
    "EndDeviceControl",
    "EndDeviceControlType",
    "EndDeviceEvent",
    "EndDeviceEventDetail",
    "EndDeviceEventType",
    "EndDeviceFunction",
    "EndDeviceGroup",
    "EndDeviceInfo",
    "EndDeviceTiming",
    "EnergyConnection",
    "EnergyConsumer",
    "EnergyConsumerPhase",
    "EnergySource",
    "EnergySourcePhase",
    "Equipment",
    "EquipmentContainer",
    "EquipmentFault",
    "Estimate",
    "Fault",
    "FaultCauseType",
    "FaultImpedance",
    "Feeder",
    "FloatQuantity",
    "FlowDirectionKind",
    "FrequencyDroopSettings",
    "FrequencyTripSettings",
    "Fuse",
    "GAPPS_CIM_export",
    "GeographicalRegion",
    "House",
    "HouseCooling",
    "HouseHeating",
    "IEEE1547AbnormalPerfomanceCategory",
    "IEEE1547ControlSettings",
    "IEEE1547Info",
    "IEEE1547IslandingCategory",
    "IEEE1547NormalPerformanceCategory",
    "IEEE1547Setting",
    "IEEE1547TripSettings",
    "IOPoint",
    "IdentifiedObject",
    "LinearShuntCompensator",
    "LinearShuntCompensatorPhase",
    "LoadBreakSwitch",
    "LoadResponseCharacteristic",
    "Location",
    "Measurement",
    "Meter",
    "MeterReading",
    "MomentaryCessationSettings",
    "Name",
    "NoLoadTest",
    "NormalOPcatKind",
    "OpenCircuitTest",
    "OperationalLimit",
    "OperationalLimitDirectionKind",
    "OperationalLimitSet",
    "OperationalLimitType",
    "OrderedPhaseCodeKind",
    "OverfrequencyTripCurve",
    "OverheadWireInfo",
    "OvervoltageTripCurve",
    "PerLengthImpedance",
    "PerLengthLineParameter",
    "PerLengthSequenceImpedance",
    "PhaseCode",
    "PhaseConnectedFaultKind",
    "PhaseShuntConnectionKind",
    "PhaseTapChanger",
    "PowerCutZone",
    "PowerElectronicsConnection",
    "PowerElectronicsConnectionPhase",
    "PowerElectronicsUnit",
    "PowerLimitSettings",
    "PowerSystemResource",
    "PowerTransformer",
    "PowerTransformerEnd",
    "PowerTransformerInfo",
    "Priority",
    "ProtectedSwitch",
    "RandomisationKind",
    "RatioTapChanger",
    "ReactiveCapabilityCurve",
    "Recloser",
    "RegularIntervalSchedule",
    "RegularTimePoint",
    "RegulatingCondEq",
    "RegulatingControl",
    "RegulatingControlModeKind",
    "RegulationSchedule",
    "RemoteInputSignal",
    "RemoteSignalKind",
    "RotatingMachine",
    "ScheduledEvent",
    "Season",
    "SeasonDayTypeSchedule",
    "Sectionaliser",
    "ServiceSettings",
    "ShortCircuitTest",
    "ShuntCompensator",
    "ShuntCompensatorInfo",
    "ShuntCompensatorPhase",
    "SinglePhaseKind",
    "StateVariable",
    "Status",
    "StringQuantity",
    "SubGeographicalRegion",
    "Substation",
    "SvEstVoltage",
    "SvInjection",
    "SvPowerFlow",
    "SvShuntCompensatorSections",
    "SvStatus",
    "SvTapStep",
    "SvVoltage",
    "Switch",
    "SwitchInfo",
    "SynchronousMachine",
    "TapChanger",
    "TapChangerControl",
    "TapChangerInfo",
    "TapSchedule",
    "TapeShieldCableInfo",
    "Terminal",
    "ThermostatControlMode",
    "ThermostatController",
    "TimeIntervalKind",
    "TopologicalIsland",
    "TopologicalNode",
    "TransformerCoreAdmittance",
    "TransformerEnd",
    "TransformerEndInfo",
    "TransformerMeshImpedance",
    "TransformerTank",
    "TransformerTankEnd",
    "TransformerTankInfo",
    "TransformerTest",
    "UnderfrequencyTripCurve",
    "UndervoltageTripCurve",
    "UnitMultiplier",
    "UnitSymbol",
    "UsagePoint",
    "UsagePointConnectedKind",
    "UsagePointGroup",
    "Version",
    "VoltVarCurve",
    "VoltVarSettings",
    "VoltWattCurve",
    "VoltWattSettings",
    "VoltageControlZone",
    "VoltageTripSettings",
    "WattVarCurve",
    "WattVarSettings",
    "WindingConnection",
    "WireAssemblyInfo",
    "WireInfo",
    "WireInsulationKind",
    "WireMaterialKind",
    "WirePhaseInfo",
    "WirePosition",
    "WireSpacingInfo",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Optional

import cimgraph.data_profile.rc4_2021 as cim


def get_all_attributes(feeder_mrid: str, typed_catalog: dict[type, dict[str, object]]) -> str: 
    """ Generates SPARQL query string for a given catalog of objects and feeder id
    Args:
        feeder_mrid (str | Feeder object): The mRID of the feeder or feeder object
        typed_catalog (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by 
            class type and object mRID
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    mrid_list = list(typed_catalog[cim.LinearShuntCompensator].keys())
    
    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <http://iec.ch/TC57/CIM100#>
        SELECT ?mRID ?name ?BaseVoltage ?Location ?RegulatingControl ?aVRDelay ?b0PerSection 
        ?bPerSection ?controlEnabled ?g0PerSection ?gPerSection ?grounded ?maximumSections ?nomU 
        ?normalSections ?phaseConnection ?sections ?Terminals ?ShuntCompensatorPhase 
        (group_concat(distinct ?Measurement; separator=";") as ?Measurements) 
        WHERE { 
          ?eq r:type cim:LinearShuntCompensator.
          VALUES ?fdrid {"%s"}
          VALUES ?mRID {"""%feeder_mrid
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += ' "%s" \n'%mrid
    # add all attributes
    query_message += """               } 
        ?eq cim:Equipment.EquipmentContainer ?fdr. #filter by feeder
        ?fdr cim:IdentifiedObject.mRID ?fdrid. #filter by feeder
        ?eq cim:IdentifiedObject.mRID ?mRID. #get cap mrid
        ?eq cim:IdentifiedObject.name ?name. #get cap name

        ?eq cim:ConductingEquipment.BaseVoltage ?bv. #get basevoltage object
        ?bv cim:IdentifiedObject.mRID ?BaseVoltage. 

        OPTIONAL {?eq cim:PowerSystemResource.Location ?loc. #get location object
            ?loc cim:IdentifiedObject.mRID ?Location.} #get location mrid

        OPTIONAL {?ctl cim:RegulatingControl.RegulatingCondEq ?eq. #get control
            ?ctl cim:RegulatingControl.enabled ?controlEnabled. #get enabled
            ?ctl cim:IdentifiedObject.mRID ?RegulatingControl} #get control mrid

        ?t cim:Terminal.ConductingEquipment ?eq.
        ?t cim:IdentifiedObject.mRID ?Terminals
        OPTIONAL {?meas cim:Measurement.Terminal ?t.
            ?meas cim:IdentifiedObject.mRID ?meas_id.
            ?meas a ?meas_cls.
            bind(concat(str(?meas_id),",",strafter(str(?meas_cls),"CIM100#")) as ?Measurement).}

        OPTIONAL {?eq cim:ShuntCompensator.grounded ?grounded.}
        OPTIONAL {?eq cim:ShuntCompensator.maximumSections ?maximumSections.}
        OPTIONAL {?eq cim:ShuntCompensator.nomU ?nomU.}
        OPTIONAL {?eq cim:ShuntCompensator.normalSections ?normalSections.}
        OPTIONAL {?eq cim:ShuntCompensator.aVRDelay ?aVRDelay.}
        OPTIONAL {?eq cim:ShuntCompensator.phaseConnection ?phaseConn.
            bind(strafter(str(?phaseConn),"CIM100#") as ?phaseConnection)}
        OPTIONAL {?spc cim:ShuntCompensatorPhase.ShuntCompensator ?eq.
            ?spc cim:IdentifiedObject.mRID ?ShuntCompensatorPhase.}

        OPTIONAL {?eq cim:LinearShuntCompensator.b0PerSection ?b0PerSection.}
        OPTIONAL {?eq cim:LinearShuntCompensator.bPerSection ?bPerSection.}
        OPTIONAL {?eq cim:LinearShuntCompensator.g0PerSection ?g0PerSection.}
        OPTIONAL {?eq cim:LinearShuntCompensator.gPerSection ?gPerSection.}
        }
        GROUP by  ?mRID ?name ?BaseVoltage ?Location ?RegulatingControl ?aVRDelay ?b0PerSection 
        ?bPerSection ?controlEnabled ?g0PerSection ?gPerSection ?grounded ?maximumSections ?nomU
        ?normalSections ?phaseConnection ?sections ?Terminals ?ShuntCompensatorPhase
        ORDER by  ?name
        """
    return query_message